      PROGRAM EULER
C
C     Programa Euler.for
C
C     Pablo Lpez Cienfuegos.
C     Jorge Lpez Garca.
C
C     Este programa resuelve numricamente un problema de Cauchy
C     bidimensional utilizando el mtodo de Euler y usando como control
C     del paso el criterio de estimacin por paso (EPP).
C
      REAL T,H,Y,Z,T1,H1,Y1,Z1,PASO,PASOINI,PASOMIN,PASOMAX,TOL,T0,TN,
     +     INICIAL,ARFS
C
C        T,H,Y,Z:       Variables que almacenan los valores Tn, Hn,
C                       Y1n, Y2n respectivamente.
C        T1,H1,Y1,Z1:   Variables que almacenan los valores Tn+1,
C                       Hn+1, Y1n+1, Y2n+1 respectivamente.
C        Paso:          Funcin que calcula el paso Hn+1 a partir de
C                       los datos del instante n.
C        Pasoini:       Variable que almacena el paso inicial.
C        Pasomin:       Almacena el paso mnimo.
C        Pasomax:       Almacena el paso mximo.
C        Tol:           Tolerancia para el error local.
C        T0:            Extremo inferior del intervalo de integracin.
C        TN:            Longitud del intervalo de integracin.
C        Inicial:       Funcin que calcula el paso inicial.
C        Arfs:          Variable que controla el tiempo de espera por
C                       la solucin.
C
      INTEGER ITERA,CAMBIO,K,NMAX,OPCION
C
C        Itera:         Numero de iteraciones realizadas.
C        Cambio:        Controla el nmero de cambios de paso
C                       realizados a lo largo del proceso.
C        K:             Dato que nos indica cada cunto hay que
C                       almacenar la aproximacin de la solucin.
C        Nmax:          Dato que nos indica el nmero mximo de
C                       iteraciones aceptable.
C        Opcion:        Variable con la que controlamos el menu de
C                       apertura del fichero de datos.
C
      CHARACTER*80 FICHERO,SALIDA
C
C        Fichero:       Nombre del fichero que contiene los datos.
C        Salida:        Nombre del fichero en el que almacenamos
C                       los datos: Ti,Hi,Y1i,Y2i.
C
      LOGICAL EXISTE
C
C        Existe:        Controla el menu de apertura del fichero de
C                       datos.
C
10    FORMAT(1X,F13.8,3(4X,F13.8))
C
C        Este es el formato con el que almacenamos los datos en el
C        Fichero de salida.
C
      EXTERNAL F1,F2
C
C        F1,F2:         Componentes de la funcin F del Problema de
C                       Valor Inicial.
C
      CALL SYSTEMQQ('CLS')
      WRITE(*,*)' '
      WRITE(*,*)'     Escriba el nombre del fichero que contiene los ',
     +          'datos para realizar'
      WRITE(*,*)'     el programa (max. 80 car.):'
      WRITE(*,*)' '
      WRITE(*,'(5X,A\)')' '
      READ(*,'(A)') FICHERO
      INQUIRE(FILE=FICHERO,EXIST=EXISTE)
      IF(.NOT.EXISTE) THEN
         CALL SYSTEMQQ('CLS')
         WRITE(*,*)' '
         WRITE(*,*)'     No existe el fichero ',FICHERO
         WRITE(*,*)'     Elija una de las siguientes opciones:'
      END IF
      DO WHILE(.NOT.EXISTE)
         WRITE(*,*)' '
         WRITE(*,*)'         1-.   Volver a introducir el nombre del ',
     +             'fichero.'
         WRITE(*,*)'         2-.   Salir del programa.'
         WRITE(*,*)' '
         WRITE(*,'(5X,A\)')' '
         READ(*,*) OPCION
         IF(OPCION.EQ.1) THEN
            CALL SYSTEMQQ('CLS')
            WRITE(*,*)' '
            WRITE(*,*)'     Introduzca el nombre del fichero que ',
     +                'contiene los datos:'
            WRITE(*,*)' '
            WRITE(*,'(5X,A\)')' '
            READ(*,'(A)') FICHERO
            INQUIRE(FILE=FICHERO,EXIST=EXISTE)
            IF(.NOT.EXISTE) THEN
               CALL SYSTEMQQ('CLS')
               WRITE(*,*)' '
               WRITE(*,*)'     No existe el fichero ',FICHERO
               WRITE(*,*)'     Elija una de las siguientes opciones:'
            END IF
         ELSE IF(OPCION.EQ.2) THEN
            STOP' '
         ELSE
            CALL SYSTEMQQ('CLS')
            WRITE(*,*)' '
            WRITE(*,*)'     Elija una de las opciones disponibles:'
         END IF
      END DO
C
C        Hemos comprobado si existe el fichero que contiene los datos.
C
      CALL SYSTEMQQ('CLS')
      WRITE(*,*)' '
      WRITE(*,*)' '
      WRITE(*,*)'     Espere un momento:'
      WRITE(*,*)' '
      WRITE(*,'(2X,A2,16X,A3,16X,A3,16X,A3,15X,A4)')'0%','25%','50%',
     +        '75%','100%'
      OPEN(1,FILE=FICHERO)
      READ(1,'(A)') SALIDA
      READ(1,*) T0
      READ(1,*) TN
      READ(1,*) Y,Z
      READ(1,*) TOL
      READ(1,*) K
      READ(1,*) NMAX
      CLOSE(1)
C
C        Hemos ledo los datos que necesita el programa.
C
      T=T0
      H=INICIAL(TOL,T,Y,Z)
      PASOINI=H
      PASOMIN=H
      PASOMAX=H
C
C        Hemos inicializado el algoritmo.
C
      OPEN(1,FILE=SALIDA)
      WRITE(1,'(5X,A,12X,A,14X,A,15X,A)') 'Tiempo','Paso','Y1','Y2'
      WRITE(1,*)'--------------   --------------   --------------   --',
     +          '------------'
      WRITE(1,10) T,H,Y,Z
      ITERA=0
      CAMBIO=0
      ARFS=T0
      WRITE(*,'(1X,\)')
C
C        Comenzamos el bucle en el que se realiza el algoritmo.
C
      DO WHILE((ITERA.NE.NMAX).AND.(H.NE.0.))
         T1=T+H
         IF(T1.LT.T0+TN) THEN
            Y1=Y+H*F1(T,Y,Z)
            Z1=Z+H*F2(T,Y,Z)
            IF(H1.GT.PASOMAX) PASOMAX=H1
            H1=PASO(H,T,Y,Z,Y1,Z1,CAMBIO,TOL)
            IF(H1.LT.PASOMIN) PASOMIN=H1
            ITERA=ITERA+1
            IF(MOD(ITERA,K).EQ.0) WRITE(1,10) T1,H1,Y1,Z1
            IF(T1.GT.ARFS) THEN
               ARFS=ARFS+TN/80.
               WRITE(*,'(A\)') ''
            END IF
            T=T1
            H=H1
            Y=Y1
            Z=Z1
         ELSE IF(T1.GE.T0+TN) THEN
            H=T0+TN-T
            IF(H.LT.PASOMIN) PASOMIN=H
            IF(H.GT.PASOMAX) PASOMAX=H
            T1=T0+TN
            Y1=Y+H*F1(T,Y,Z)
            Z1=Z+H*F2(T,Y,Z)
            H1=0.
            ITERA=ITERA+1
            WRITE(1,10) T1,H1,Y1,Z1
            T=T1
            H=H1
            Y=Y1
            Z=Z1
         END IF
      END DO
C
C        Este bucle tiene dos tests de parada; uno comprueba que no se
C        supera el nmero mximo de iteraciones comunicado en el fichero
C        de datos, y el otro es el que se encarga de comprobar que al
C        calcular el nuevo paso no se han superado los 6 cambios de
C        paso. (Cuando se superan los 6 cambios de paso, la funcin
C        Paso devuelve un valor cero).
C
      CALL SYSTEMQQ('CLS')
      IF((T.LT.T0+TN).AND.(H.EQ.0.)) THEN
         WRITE(*,*)' '
         WRITE(1,*)'Se ha alcanzado el numero maximo de cambios de paso'
         WRITE(*,*)'Se ha alcanzado el numero maximo de cambios de paso'
      ELSE IF((T.LT.T0+TN).AND.(ITERA.EQ.NMAX)) THEN
         WRITE(*,*)' '
         WRITE(1,*)'Se ha superado el numero maximo de iteraciones'
         WRITE(*,*)'Se ha superado el numero maximo de iteraciones; ',
     +             'intente volver a resolver'
         WRITE(*,*)'el problema con una tolerancia mayor.'
      END IF
      CLOSE(1)
C
C        Comunicamos los resultados por pantalla.
C
      IF(T.EQ.T0+TN) THEN
         WRITE(*,*) ' '
         WRITE(*,*) '      Valor aproximado de Y1 en',T0+TN,' ........',
     +               Y
         WRITE(*,*) '      Valor aproximado de Y2 en',T0+TN,' ........',
     +               Z
         WRITE(*,*) '      Iteraciones realizadas ....................',
     +              '......',ITERA
         WRITE(*,*) '      Numero de cambios de paso .................',
     +              '......',CAMBIO
         WRITE(*,*) '      Paso inicial ..............................',
     +              '......',PASOINI
         WRITE(*,*) '      Paso minimo ...............................',
     +              '......',PASOMIN
         WRITE(*,*) '      Paso maximo ...............................',
     +              '......',PASOMAX
      END IF
      END

C-----------------------------------------------------------------------
      FUNCTION INICIAL(TOL,T0,Y0,Z0)
C
C        Funcin que calcula el paso inicial.
C
      REAL FF1,FF2,TOL,T0,Y0,Z0,INICIAL,NORMA
C
C        FF1:           Almacena el valor de la primera componente
C                       de F en el instante inicial.
C        FF2:           Almacena el valor de la segunda componente
C                       de F en el instante inicial.
C        Tol:           Tolerancia para el error local.
C        T0,Y0,Z0:      Valores iniciales de Tn, Y1n, Y2n.
C        Inicial:       Variable que almacena el paso inicial.
C        Norma:         Funcin que calcula la norma eucldea.
C
      EXTERNAL F1,F2
C
C        F1,F2:         Componentes de la funcin F del Problema de
C                       Valor Inicial.
C
      FF1=F1(T0,Y0,Z0)
      FF2=F2(T0,Y0,Z0)
      IF ((FF1.EQ.0.).AND.(FF2.EQ.0.)) THEN
         INICIAL=SQRT(TOL)
      ELSE
         INICIAL=MIN(SQRT(TOL/NORMA(FF1,FF2)),SQRT(TOL))
      END IF
      RETURN
      END

C-----------------------------------------------------------------------
      FUNCTION EPS(H,T,Y,Z,Y1,Z1)
C
C        Funcin que calcula la aproximacin numrica del error local
C        en la etapa n-sima.
C
      REAL H,T,Y,Z,Y1,Z1,EPS,NORMA
C
C        H,T,Y,Z:       Hn, Tn, Y1n, Y2n respectivamente.
C        Y1,Z1:         Y1n+1, Y2n+1 respectivamente.
C        Eps:           Almacena la aproximacin numrica del error
C                       local en la etapa n-sima.
C        Norma:         Funcin que calcula la norma eucldea.
C
      EXTERNAL F1,F2
C
C        F1,F2:         Componentes de la funcin F del Problema de
C                       Valor Inicial.
C
      EPS=H*.5*NORMA(F1(T+H,Y1,Z1)-F1(T,Y,Z),F2(T+H,Y1,Z1)-F2(T,Y,Z))
      RETURN
      END

C-----------------------------------------------------------------------
      FUNCTION PASO(H,T,Y,Z,Y1,Z1,CAMBIO,TOL)
C
C        Funcin que calcula el paso Hn+1.
C
      REAL H,T,Y,Z,Y1,Z1,E,PASO,EPS,TOL
C
C        H,T,Y,Z:       Hn, Tn, Y1n, Y2n respectivamente.
C        Y1,Z1:         Y1n+1, Y2n+1 respectivamente.
C        E:             Almacena el valor absoluto de la aproximacin
C                       numrica del error local en el paso n-esimo.
C        Paso:          Almacena el nuevo paso.
C        Eps:           Funcin que calcula la aproximacin numrica
C                       del error local en el paso n-simo.
C        Tol:           Tolerancia para el error local.
C
      INTEGER CAMBIO,RECHAZO
C
C        Cambio:        Almacena el nmero de cambios de paso
C                       realizados a lo largo del proceso.
C        Rechazo:       Almacena el nmero de cambios de paso
C                       que se han realizado en la etapa actual del
C                       proceso.
C
      E=ABS(EPS(H,T,Y,Z,Y1,Z1))
      IF (E.EQ.0) THEN
         H=H*2.
      ELSE
         H=H*.85*SQRT(TOL/E)
      END IF
      E=ABS(EPS(H,T,Y,Z,Y1,Z1))
      RECHAZO=0
C
C        Si se rechaza el paso propuesto, entramos en el siguiente bucle
C        para intentar encontrar el nuevo paso.
C
      IF (E.GE.TOL) CAMBIO=CAMBIO+1
      DO WHILE ((E.GE.TOL).AND.(RECHAZO.LT.6))
         RECHAZO=RECHAZO+1
         H=H*.85*SQRT(TOL/E)
         E=ABS(EPS(H,T,Y,Z,Y1,Z1))
      END DO
      IF (E.GE.TOL) THEN
         PASO=0
      ELSE
         PASO=H
      END IF
C
C        Si no podemos encontrar un nuevo paso vlido tras cambiar 6
C        veces de paso, le devolvemos al programa que nos ha llamado
C        un nuevo paso con valor cero, para gestionar la finalizacin
C        del programa desde el programa que nos llama.
C
      RETURN
      END

C-----------------------------------------------------------------------
      FUNCTION NORMA(NUM1,NUM2)
C
C        Funcin que calcula la norma eucldea.
C
      REAL NORMA,NUM1,NUM2
C
C        Norma:         Variable que contiene el valor de la norma.
C        Num1,Num2:     Variables auxiliares.
C
      NORMA=(NUM1**2+NUM2**2)**.5
      RETURN
      END
