C-----------------------------------------------------------------------
C                               OPCION A
C-----------------------------------------------------------------------
      PROGRAM OPCION_A
C
C     Programa Ppal_a.for
C
C     Pablo Lpez Cienfuegos.
C     Jorge Lpez Garca.
C
C     Este programa resuelve numricamente un problema de Cauchy
C     bidimensional utilizando dos pares de mtodos encajados de 
C     Fehlberg RKF4(5) y usando como control de paso el criterio de
C     estimacin por paso (EPP).
C
C     Para cambiar el nmero de etapas del mtodo, basta con modificar
C     las instrucciones PARAMETER que aparecen a lo largo del programa.
C
      INTEGER ITERA,CAMBIO,K,NMAX,OPCION,DIM
C
C        Itera:         Numero de iteraciones realizadas.
C        Cambio:        Controla el nmero de cambios de paso
C                       realizados a lo largo del proceso.
C        K:             Dato que nos indica cada cunto hay que
C                       almacenar la aproximacin de la solucin.
C        Nmax:          Dato que nos indica el nmero mximo de
C                       iteraciones aceptable.
C        Opcion:        Variable con la que controlamos el menu de
C                       apertura del fichero de datos.
C        Dim:           Nmero de etapas del segundo mtodo.
C
      PARAMETER(DIM=6)
      REAL T,H,Y,Z,T1,H1,Y1,Z1,PASO_A,PASOINI,PASOMIN,PASOMAX,TOL,T0,TN,
     +     INICIAL_A,MDC,A(1:DIM,1:DIM),B2(1:DIM),B1(1:DIM),C(1:DIM),
     +     K1(1:DIM),K2(1:DIM),PARTE1,PARTE2
C
C        T,H,Y,Z:       Variables que almacenan los valores Tn, Hn,
C                       Y1n,Y2n respectivamente.
C        T1,H1,Y1,Z1:   Variables que almacenan los valores Tn+1,
C                       Hn+1, Y1n+1, Y2n+1 respectivamente.
C        Paso_A:        Funcin que calcula el paso Hn+1 a partir de
C                       los datos del instante n.
C        Pasoini:       Variable que almacena el paso inicial.
C        Pasomin:       Almacena el paso mnimo.
C        Pasomax:       Almacena el paso mximo.
C        Tol:           Tolerancia para el error local.
C        T0:            Extremo inferior del intervalo de
C                       integracin.
C        TN:            Longitud del intervalo de integracin.
C        Inicial_A:     Funcin que calcula el paso inicial.
C        MdC:           Variable que controla el tiempo de espera por
C                       la solucin.
C        A(i,j):        Coeficientes de la matriz A.
C        B1(i):         Coeficientes del vector B1.
C        B2(i):         Coeficientes del vector B2.
C        C(i):          Coeficientes del vector C.
C        Kj(i):         Valores de las K's del mtodo.
C        Partei:        Variables auxiliares.
C        Norma:         Variable que contiene el valor de la norma 
C                       eucldea de dos nmeros reales.
C        Num1,Num2:     Variables auxiliares.
C
      CHARACTER*80 FICHERO,FICHERO1
C
C        Fichero:       Nombre del fichero que contiene los datos.
C        Fichero1:      Nombre del fichero en el que almacenamos
C                       los datos: Ti,Hi,Y1i,Y2i.
C
      LOGICAL EXISTE
C
C        Existe:        Controla el menu de apertura del fichero de
C                       datos.
C
10    FORMAT(1X,F13.8,3(4X,F13.8))
C
C        Este es el formato con el que almacenamos los datos en el
C        Fichero1.
C
      EXTERNAL F1,F2
C
C        F1,F2:         Componentes de la funcin F del Problema de
C                       Valor Inicial.
C
      CALL SYSTEMQQ('CLS')
      WRITE(*,*)' '
      WRITE(*,*)'     Escriba el nombre del fichero que contiene los ',
     +          'datos para realizar'
      WRITE(*,*)'     el programa (max. 80 car.):'
      WRITE(*,*)' '
      WRITE(*,'(5X,A\)')' '
      READ(*,'(A)') FICHERO
      INQUIRE(FILE=FICHERO,EXIST=EXISTE)
      IF(.NOT.EXISTE) THEN
         CALL SYSTEMQQ('CLS')
         WRITE(*,*)' '
         WRITE(*,*)'     No existe el fichero ',FICHERO
         WRITE(*,*)'     Elija una de las siguientes opciones:'
      END IF
      DO WHILE(.NOT.EXISTE)
         WRITE(*,*)' '
         WRITE(*,*)'         1-.   Volver a introducir el nombre del ',
     +             'fichero.'
         WRITE(*,*)'         2-.   Salir del programa.'
         WRITE(*,*)' '
         WRITE(*,'(5X,A\)')' '
         READ(*,*) OPCION
         IF(OPCION.EQ.1) THEN
            CALL SYSTEMQQ('CLS')
            WRITE(*,*)' '
            WRITE(*,*)'     Introduzca el nombre del fichero que ',
     +                'contiene los datos:'
            WRITE(*,*)' '
            WRITE(*,'(5X,A\)')' '
            READ(*,'(A)') FICHERO

            INQUIRE(FILE=FICHERO,EXIST=EXISTE)
            IF(.NOT.EXISTE) THEN
               CALL SYSTEMQQ('CLS')
               WRITE(*,*)' '
               WRITE(*,*)'     No existe el fichero ',FICHERO
               WRITE(*,*)'     Elija una de las siguientes opciones:'
            END IF
         ELSE IF(OPCION.EQ.2) THEN
            STOP' '
         ELSE
            CALL SYSTEMQQ('CLS')
            WRITE(*,*)' '
            WRITE(*,*)'     Elija una de las opciones disponibles:'
         END IF
      END DO
C
C        Hemos comprobado si existe el fichero que contiene los datos.
C
      CALL SYSTEMQQ('CLS')
      WRITE(*,*)' '
      WRITE(*,*)'     Espere un momento:'
      WRITE(*,*)' '
      WRITE(*,'(2X,A2,16X,A3,16X,A3,16X,A3,15X,A4)')'0%','25%','50%',
     +        '75%','100%'
      OPEN(1,FILE=FICHERO)
      READ(1,'(A)') FICHERO1
      READ(1,*) T0
      READ(1,*) TN
      READ(1,*) Y,Z
      READ(1,*) TOL
      READ(1,*) K
      READ(1,*) NMAX
      READ(1,*) ((A(I,J), J = 1,DIM), I =1,DIM)
      READ(1,*) (C(J),J=1,DIM)
      READ(1,*) (B1(J),J=1,DIM-1)
      READ(1,*) (B2(J),J=1,DIM)
      CLOSE(1)
C
C        Hemos ledo los datos que necesita el programa.
C
      B1(DIM)=0.
      T=T0
      H=INICIAL_A(TOL,T,Y,Z)
      PASOINI=H
      PASOMIN=H
      PASOMAX=H
C
C        Hemos inicializado el algoritmo.
C
      OPEN(1,FILE=FICHERO1)
      WRITE(1,'(5X,A,12X,A,14X,A,15X,A)') 'Tiempo','Paso','Y1','Y2'
      WRITE(1,*)'--------------   --------------   --------------   --',
     +          '------------'
      WRITE(1,10) T,H,Y,Z
      ITERA=0
      CAMBIO=0
      MDC=T0
      WRITE(*,'(1X,\)')
C
C        Comenzamos el bucle en el que se realiza el algoritmo.
C
      DO WHILE((ITERA.NE.NMAX).AND.(H.NE.0.))
         T1=T+H
         IF(T1.LT.T0+TN) THEN
            IF(H.GT.PASOMAX) PASOMAX=H
            IF(H.LT.PASOMIN) PASOMIN=H
            H1=PASO_A(H,T,Y,Z,CAMBIO,TOL,A,B2,B1,C,K1,K2)
            PARTE1=0.
            PARTE2=0.
            DO I=1,DIM-1
               PARTE1=PARTE1+B1(I)*K1(I)
               PARTE2=PARTE2+B1(I)*K2(I)
            END DO
            Y1=Y+H*PARTE1
            Z1=Z+H*PARTE2
            ITERA=ITERA+1
            IF(MOD(ITERA,K).EQ.0) THEN
               WRITE(1,10) T1,H1,Y1,Z1
            END IF
            IF(T1.GT.MDC) THEN
               MDC=MDC+TN/80.
               WRITE(*,'(A\)') ''
            END IF
            T=T1
            H=H1
            Y=Y1
            Z=Z1
         ELSE IF(T1.GE.T0+TN) THEN
            H=T0+TN-T
            IF(H.LT.PASOMIN) PASOMIN=H
            IF(H.GT.PASOMAX) PASOMAX=H
            T1=T0+TN
            CALL CALCULAR_K(H,T,Y,Z,A,C,K1,K2)
            PARTE1=0.
            PARTE2=0.
            DO I=1,DIM-1
               PARTE1=PARTE1+B1(I)*K1(I)
               PARTE2=PARTE2+B1(I)*K2(I)
            END DO
            Y1=Y+H*PARTE1
            Z1=Z+H*PARTE2
            H1=0.
            ITERA=ITERA+1
            WRITE(1,10) T1,H1,Y1,Z1
            T=T1
            H=H1
            Y=Y1
            Z=Z1
         END IF
      END DO
C
C        Este bucle tiene dos tests de parada; uno comprueba que no se
C        supera el nmero mximo de iteraciones comunicado en el fichero
C        de datos, y el otro es el que se encarga de comprobar que al
C        calcular el nuevo paso no se han superado los 6 cambios de
C        paso. (Cuando se superan los 6 cambios de paso, la funcin
C        Paso devuelve un valor cero).
C
      CALL SYSTEMQQ('CLS')
      IF((T.LT.T0+TN).AND.(H.EQ.0.)) THEN
         WRITE(*,*)' '
         WRITE(1,*)'Se ha alcanzado el numero maximo de cambios de paso'
         WRITE(*,*)'Se ha alcanzado el numero maximo de cambios de paso'
      ELSE IF((T.LT.T0+TN).AND.(ITERA.EQ.NMAX)) THEN
         WRITE(*,*)' '
         WRITE(1,*)'Se ha superado el numero maximo de iteraciones'
         WRITE(*,*)'Se ha superado el numero maximo de iteraciones; ',
     +             'intente volver a resolver'
      WRITE(*,*)'el problema con una tolerancia mayor.'
      END IF
      CLOSE(1)
C
C        Comunicamos los resultados por pantalla.
C
      IF(T.EQ.T0+TN) THEN
         WRITE(*,*) ' '
         WRITE(*,*) '      Valor aproximado de Y1 en',T0+TN,' ........',
     +               Y
         WRITE(*,*) '      Valor aproximado de Y2 en',T0+TN,' ........',
     +               Z
         WRITE(*,*) '      Iteraciones realizadas ....................',
     +              '......',ITERA
         WRITE(*,*) '      Numero de cambios de paso .................',
     +              '......',CAMBIO
         WRITE(*,*) '      Paso inicial ..............................',
     +              '......',PASOINI
         WRITE(*,*) '      Paso minimo ...............................',
     +              '......',PASOMIN
         WRITE(*,*) '      Paso maximo ...............................',
     +              '......',PASOMAX
      END IF
      END

C-----------------------------------------------------------------------
      FUNCTION INICIAL_A(TOL,T0,Y0,Z0)
C
C        Funcin que calcula el paso inicial.
C
      REAL FF1,FF2,TOL,T0,Y0,Z0,INICIAL_A,NORMA
C
C        FF1:           Almacena el valor de la primera componente
C                       de F en el instante inicial.
C        FF2:           Almacena el valor de la segunda componente
C                       de F en el instante inicial.
C        Tol:           Tolerancia para el error local.
C        T0,Y0,Z0:      Valores iniciales de Tn, Y1n, Y2n.
C        Inicial_A:     Variable que almacena el paso inicial.
C        Norma:         Funcin que calcula la norma de dos nmeros.
C
      EXTERNAL F1,F2
C
C        F1,F2:         Componentes de la funcin F del Problema de
C                       Valor Inicial.
C
      FF1=F1(T0,Y0,Z0)
      FF2=F2(T0,Y0,Z0)
      IF ((FF1.EQ.0.).AND.(FF2.EQ.0.)) THEN
         INICIAL_A=TOL**.2
      ELSE
         INICIAL_A=(TOL/NORMA(FF1,FF2))**.2
      END IF
      RETURN
      END

C-----------------------------------------------------------------------
      SUBROUTINE CALCULAR_K(H,T,Y,Z,A,C,K1,K2)
C
C        Funcin que calcula los coeficientes K en la etapa n-sima.
C
      INTEGER DIM
C
C        Dim:           Nmero de etapas del segundo mtodo.
C
      PARAMETER(DIM=6)
      REAL T,H,Y,Z,A(1:DIM,1:DIM),C(1:DIM),K1(1:DIM),K2(1:DIM)
C
C        T,H,Y,Z:       Variables que almacenan los valores Tn, Hn,
C                       Y1n,Y2n respectivamente.
C        A(i,j):        Coeficientes de la matriz A.
C        C(i):          Coeficientes del vector C.
C        Kj(i):         Valores de las K's del mtodo.
C
      EXTERNAL F1,F2
C
C        F1,F2:         Componentes de la funcin F del Problema de
C                       Valor Inicial.
C
      K1(1)=F1(T,Y,Z)
      K2(1)=F2(T,Y,Z)
C
      DO I=2,DIM
         SUMA1=0.
         SUMA2=0.
         DO J=1,I-1
            SUMA1=SUMA1+A(I,J)*K1(J)
            SUMA2=SUMA2+A(I,J)*K2(J)
         END DO
         K1(I)=F1(T+C(I)*H,Y+H*SUMA1,Z+H*SUMA2)
         K2(I)=F2(T+C(I)*H,Y+H*SUMA1,Z+H*SUMA2)
      END DO
C
      RETURN
      END

C-----------------------------------------------------------------------
      FUNCTION EPS_A(H,T,Y,Z,A,B2,B1,C,K1,K2)
C
C        Funcin que calcula la aproximacin numrica del error local
C        en la etapa n-sima.
C
      INTEGER DIM
C
C        Dim:           Nmero de etapas del segundo mtodo.
C
      PARAMETER(DIM=6)
      REAL EPS_A,EPS1,EPS2,T,H,Y,Z,A(1:DIM,1:DIM),B2(1:DIM),B1(1:DIM),
     +     C(1:DIM),K1(1:DIM),K2(1:DIM),NORMA
C
C        Eps_A:         Almacena la aproximacin numrica del error
C                       local en la etapa n-sima.
C        Epsi:          Variables auxiliares.
C        T,H,Y,Z:       Variables que almacenan los valores Tn, Hn,
C                       Y1n,Y2n respectivamente.
C        A(i,j):        Coeficientes de la matriz A.
C        B1(i):         Coeficientes del vector B1.
C        B2(i):         Coeficientes del vector B2.
C        C(i):          Coeficientes del vector C.
C        Kj(i):         Valores de las K's del mtodo.
C        Norma:         Funcin que calcula la norma de dos nmeros.
C
      EXTERNAL F1,F2
C
C        F1,F2:         Componentes de la funcin F del Problema de
C                       Valor Inicial.
C
      CALL CALCULAR_K(H,T,Y,Z,A,C,K1,K2)
C
      EPS1=0.
      EPS2=0.
      DO I=1,DIM
         EPS1=EPS1+(B2(I)-B1(I))*K1(I)
         EPS2=EPS2+(B2(I)-B1(I))*K2(I)
      END DO
      EPS_A=H*NORMA(EPS1,EPS2)
      RETURN
      END

C-----------------------------------------------------------------------
      FUNCTION PASO_A(H,T,Y,Z,CAMBIO,TOL,A,B2,B1,C,K1,K2)
C
C        Funcin que calcula el paso Hn+1
C
      INTEGER CAMBIO,RECHAZO,DIM
C
C        Cambio:        Almacena el nmero de cambios de paso
C                       realizados a lo largo del proceso.

C        Rechazo:       Almacena el nmero de cambios de paso
C                       que se han realizado en la etapa actual del
C                       proceso.
C        Dim:           Nmero de etapas del segundo mtodo.
C
      PARAMETER(DIM=6)
      REAL H,T,Y,Z,E,PASO_A,EPS_A,TOL,A(1:DIM,1:DIM),B2(1:DIM),
     +     B1(1:DIM),C(1:DIM),K1(1:DIM),K2(1:DIM)
C
C        H,T,Y,Z:       Hn, Tn, Y1n, Y2n respectivamente.
C        E:             Almacena el valor absoluto de la aproximacin
C                       numrica del error local en el paso n-esimo.
C        Paso_A:        Almacena el nuevo paso.
C        Eps_A:         Funcin que calcula la aproximacin numrica
C                       del error local en el paso n-simo.
C        Tol:           Tolerancia para el error local.
C        A(i,j):        Coeficientes de la matriz A.
C        B1(i):         Coeficientes del vector B1.
C        B2(i):         Coeficientes del vector B2.
C        C(i):          Coeficientes del vector C.
C        Kj(i):         Valores de las K's del mtodo.
C
      PASO_A=H
      E=EPS_A(PASO_A,T,Y,Z,A,B2,B1,C,K1,K2)
      IF(E.EQ.0.) THEN
          PASO_A=PASO_A*2.
      ELSE
          PASO_A=PASO_A*.85*((TOL/E)**.2)
      END IF
      E=EPS_A(PASO_A,T,Y,Z,A,B2,B1,C,K1,K2)
      RECHAZO=0
C
C        Si se rechaza el paso propuesto, entramos en el siguiente bucle
C        para intentar encontrar el nuevo paso.
C
      IF (E.GE.TOL) CAMBIO=CAMBIO+1
      DO WHILE ((E.GE.TOL).AND.(RECHAZO.LT.6))
         RECHAZO=RECHAZO+1
         PASO_A=PASO_A*.85*((TOL/E)**.2)
         E=EPS_A(PASO_A,T,Y,Z,A,B2,B1,C,K1,K2)
      END DO
      IF (E.GE.TOL) PASO_A=0.
C
C        Si no podemos encontrar un nuevo paso vlido tras cambiar 6
C        veces de paso, le devolvemos al programa que nos ha llamado
C        un nuevo paso con valor cero, para gestionar la finalizacin
C        del programa desde el programa que nos llama.
C
      RETURN
      END
       
C-----------------------------------------------------------------------
      FUNCTION NORMA(NUM1,NUM2)
C
C        Funcin que calcula la norma eucldea.
C
      REAL NORMA,NUM1,NUM2
C

C        Norma:         Variable que contiene el valor de la norma.
C        Num1,Num2:     Variables auxiliares.
C
      NORMA=(NUM1**2.+NUM2**2.)**.5
      RETURN
      END       